@@ -1,25 +1,26 @@
 use std::collections::HashMap;
-use std::fmt;
 use std::io::{ErrorKind};
-use std::net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr};
+use std::net::{IpAddr, Ipv4Addr, SocketAddr};
 use tokio::net::UdpSocket;
 use std::time::Duration;
 
+
 use log::{debug, info};
-use stun::addr::MappedAddress;
-use stun::agent::TransactionId;
-use stun::attributes::{ATTR_CHANGE_REQUEST, ATTR_CHANGED_ADDRESS, ATTR_MAPPED_ADDRESS, AttrType, RawAttribute};
-use stun::Error;
-use stun::message::{CLASS_ERROR_RESPONSE, Message, MessageClass, MessageType, METHOD_BINDING};
+use bytecodec::{DecodeExt, EncodeExt};
+use stun_codec_blazh::{Attribute, Message, MessageClass, MessageDecoder, MessageEncoder, TransactionId};
+use stun_codec_blazh::rfc3489::attributes::ChangedAddress;
+use stun_codec_blazh::rfc5389::attributes::{MappedAddress, UnknownAttributes};
+use stun_codec_blazh::rfc5389::methods;
+use stun_codec_blazh::rfc5780::attributes::ChangeRequest;
 use crate::NatType::{FullCone, OpenInternet, PortRestrictedCone, RestrictedCone, Symmetric, SymmetricUdpFirewall, Unknown};
 
 
 type IoResult<T>= std::io::Result<T>;
 
-pub(crate) const FAMILY_IPV4: u16 = 0x01;
-pub(crate) const FAMILY_IPV6: u16 = 0x02;
-pub(crate) const IPV4LEN: usize = 4;
-pub(crate) const IPV6LEN: usize = 16;
+// pub(crate) const FAMILY_IPV4: u16 = 0x01;
+// pub(crate) const FAMILY_IPV6: u16 = 0x02;
+// pub(crate) const IPV4LEN: usize = 4;
+// pub(crate) const IPV6LEN: usize = 16;
 
 pub const TIMEOUT: Duration = Duration::from_millis(1000);
 
@@ -191,34 +192,34 @@ fn empty_address() -> SocketAddr{
     SocketAddr::new(IpAddr::V4(Ipv4Addr::from(0)), 0)
 }
 
-#[derive(Debug)]
-pub struct ChangedAddress {
-    pub ip: IpAddr,
-    pub port: u16,
-}
-
-impl Default for ChangedAddress {
-    fn default() -> Self {
-        ChangedAddress {
-            ip: IpAddr::V4(Ipv4Addr::from(0)),
-            port: 0,
-        }
-    }
-}
-
-impl fmt::Display for ChangedAddress {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        let family = match self.ip {
-            IpAddr::V4(_) => FAMILY_IPV4,
-            IpAddr::V6(_) => FAMILY_IPV6,
-        };
-        if family == FAMILY_IPV4 {
-            write!(f, "{}:{}", self.ip, self.port)
-        } else {
-            write!(f, "[{}]:{}", self.ip, self.port)
-        }
-    }
-}
+// #[derive(Debug)]
+// pub struct ChangedAddress {
+//     pub ip: IpAddr,
+//     pub port: u16,
+// }
+//
+// impl Default for ChangedAddress {
+//     fn default() -> Self {
+//         ChangedAddress {
+//             ip: IpAddr::V4(Ipv4Addr::from(0)),
+//             port: 0,
+//         }
+//     }
+// }
+//
+// impl fmt::Display for ChangedAddress {
+//     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+//         let family = match self.ip {
+//             IpAddr::V4(_) => FAMILY_IPV4,
+//             IpAddr::V6(_) => FAMILY_IPV6,
+//         };
+//         if family == FAMILY_IPV4 {
+//             write!(f, "{}:{}", self.ip, self.port)
+//         } else {
+//             write!(f, "[{}]:{}", self.ip, self.port)
+//         }
+//     }
+// }
 
 pub async fn local_ip() -> IoResult<SocketAddr> {
     let socket = UdpSocket::bind("0.0.0.0:0").await?;
@@ -226,60 +227,61 @@ pub async fn local_ip() -> IoResult<SocketAddr> {
     socket.local_addr()
 }
 
-impl ChangedAddress {
-    /// get_from_as decodes MAPPED-ADDRESS value in message m as an attribute of type t.
-    pub fn get_from_as(&mut self, m: &Message, t: AttrType) -> std::result::Result<(), Error> {
-        let v = m.get(t)?;
-        if v.len() <= 4 {
-            return Err(Error::ErrUnexpectedEof);
-        }
-
-        let family = u16::from_be_bytes([v[0], v[1]]);
-        if family != FAMILY_IPV6 && family != FAMILY_IPV4 {
-            return Err(Error::Other(format!("bad value {}", family)));
-        }
-        self.port = u16::from_be_bytes([v[2], v[3]]);
-
-        if family == FAMILY_IPV6 {
-            let mut ip = [0; IPV6LEN];
-            let l = std::cmp::min(ip.len(), v[4..].len());
-            ip[..l].copy_from_slice(&v[4..4 + l]);
-            self.ip = IpAddr::V6(Ipv6Addr::from(ip));
-        } else {
-            let mut ip = [0; IPV4LEN];
-            let l = std::cmp::min(ip.len(), v[4..].len());
-            ip[..l].copy_from_slice(&v[4..4 + l]);
-            self.ip = IpAddr::V4(Ipv4Addr::from(ip));
-        };
-
-        Ok(())
-    }
-
-    /// add_to_as adds MAPPED-ADDRESS value to m as t attribute.
-    pub fn add_to_as(&self, m: &mut Message, t: AttrType) -> std::result::Result<(), Error> {
-        let family = match self.ip {
-            IpAddr::V4(_) => FAMILY_IPV4,
-            IpAddr::V6(_) => FAMILY_IPV6,
-        };
-
-        let mut value = vec![0u8; 4];
-        //value[0] = 0 // first 8 bits are zeroes
-        value[0..2].copy_from_slice(&family.to_be_bytes());
-        value[2..4].copy_from_slice(&self.port.to_be_bytes());
-
-        match self.ip {
-            IpAddr::V4(ipv4) => value.extend_from_slice(&ipv4.octets()),
-            IpAddr::V6(ipv6) => value.extend_from_slice(&ipv6.octets()),
-        };
-
-        m.add(t, &value);
-        Ok(())
-    }
-}
+// impl ChangedAddress {
+//     /// get_from_as decodes MAPPED-ADDRESS value in message m as an attribute of type t.
+//     pub fn get_from_as(&mut self, m: &Message, t: AttrType) -> std::result::Result<(), Error> {
+//         let v = m.get(t)?;
+//         if v.len() <= 4 {
+//             return Err(Error::ErrUnexpectedEof);
+//         }
+//
+//         let family = u16::from_be_bytes([v[0], v[1]]);
+//         if family != FAMILY_IPV6 && family != FAMILY_IPV4 {
+//             return Err(Error::Other(format!("bad value {}", family)));
+//         }
+//         self.port = u16::from_be_bytes([v[2], v[3]]);
+//
+//         if family == FAMILY_IPV6 {
+//             let mut ip = [0; IPV6LEN];
+//             let l = std::cmp::min(ip.len(), v[4..].len());
+//             ip[..l].copy_from_slice(&v[4..4 + l]);
+//             self.ip = IpAddr::V6(Ipv6Addr::from(ip));
+//         } else {
+//             let mut ip = [0; IPV4LEN];
+//             let l = std::cmp::min(ip.len(), v[4..].len());
+//             ip[..l].copy_from_slice(&v[4..4 + l]);
+//             self.ip = IpAddr::V4(Ipv4Addr::from(ip));
+//         };
+//
+//         Ok(())
+//     }
+//
+//     /// add_to_as adds MAPPED-ADDRESS value to m as t attribute.
+//     pub fn add_to_as(&self, m: &mut Message, t: AttrType) -> std::result::Result<(), Error> {
+//         let family = match self.ip {
+//             IpAddr::V4(_) => FAMILY_IPV4,
+//             IpAddr::V6(_) => FAMILY_IPV6,
+//         };
+//
+//         let mut value = vec![0u8; 4];
+//         //value[0] = 0 // first 8 bits are zeroes
+//         value[0..2].copy_from_slice(&family.to_be_bytes());
+//         value[2..4].copy_from_slice(&self.port.to_be_bytes());
+//
+//         match self.ip {
+//             IpAddr::V4(ipv4) => value.extend_from_slice(&ipv4.octets()),
+//             IpAddr::V6(ipv6) => value.extend_from_slice(&ipv6.octets()),
+//         };
+//
+//         m.add(t, &value);
+//         Ok(())
+//     }
+// }
 
 
 pub async fn nat_detect(local_address: SocketAddr,stun_server: &str) -> IoResult<(String, SocketAddr, NatType)> {
-    let transaction_id = TransactionId::new();
+
+    let transaction_id = TransactionId::new([3; 12]);
 
     let mut socket = tokio::net::UdpSocket::bind(format!("{}:0", local_address.ip())).await?;
     let mut_socket_ref = &mut socket;
@@ -287,76 +289,72 @@ pub async fn nat_detect(local_address: SocketAddr,stun_server: &str) -> IoResult
     let stun_server_string = stun_server.to_string();
 
     // test1
-    let test1_message = build_request_bind_message_with_attribute(
-        transaction_id, RawAttribute{
-            typ: ATTR_CHANGE_REQUEST,
-            value: vec![0, 0, 0, 0b0000000],
-            length: 4
-        }
-    );
-    debug!("[{}] test1 send: {}", stun_server, test1_message);
+    let test1_message: Message<UnknownAttributes> = build_request_bind_message(transaction_id);
+    debug!("[{}] test1 send: {:?}", stun_server, test1_message);
     let result = single_send(stun_server, test1_message, mut_socket_ref).await;
     debug!("[{}] test1: {}", stun_server, result.is_ok());
     if result.is_err() {
-        return IoResult::Ok((stun_server_string, SocketAddr::new(
-            IpAddr::V4(Ipv4Addr::from(0)), 0) ,NatType::UdpBlocked)
-        );
+        return IoResult::Ok((stun_server_string, empty_address(),NatType::UdpBlocked));
     }
-    let test1_response: Message = result.unwrap();
-    debug!("[{}] test1 recv: {}", stun_server, test1_response);
-    let mut mapped_address = MappedAddress::default();
-    let _ = mapped_address.get_from_as(& test1_response, ATTR_MAPPED_ADDRESS);
-    let test1_mapped_address = mapped_address;
+    let test1_response: Message<stun_codec_blazh::rfc5389::Attribute> = result.unwrap();
+    debug!("[{}] test1 recv: {:?}", stun_server, test1_response);
+
+    let test1_mapped_address = {
+        let opt: Option<&MappedAddress> = test1_response.get_attribute();
+        match opt {
+            None => return other_error(),
+            Some(a) => a.address()
+        }
+    };
     debug!("[{}] test1 mapped_address: {}", stun_server,test1_mapped_address);
 
-    let public_address = SocketAddr::new(test1_mapped_address.ip.clone(), test1_mapped_address.port.clone());
+    let public_address = SocketAddr::new(test1_mapped_address.ip().clone(), test1_mapped_address.port().clone());
 
-
-    let mut changed_address = ChangedAddress::default();
-    let _ = changed_address.get_from_as(& test1_response, ATTR_CHANGED_ADDRESS);
-    let test1_changed_address = changed_address;
+    let test1_changed_address ={
+        let opt: Option<&ChangedAddress> = test1_response.get_attribute();
+        match opt {
+            None => return other_error(),
+            Some(a) => a.address()
+        }
+    };
     debug!("[{}] test1 changed_address: {}", stun_server,test1_changed_address);
 
 
     // test2
     let test2_message = build_request_bind_message_with_attribute(
-        transaction_id, RawAttribute{
-            typ: ATTR_CHANGE_REQUEST,
-            value: vec![0, 0, 0, 0b00000110],
-            length: 4
-        }
+        transaction_id, ChangeRequest::new(true, true)
     );
 
     let local_ip = mut_socket_ref.local_addr()?.ip();
-    let test1_mapped_address_ip = test1_mapped_address.ip;
+    let test1_mapped_address_ip = test1_mapped_address.ip();
     let test1_is_same_ip = local_ip.eq(&test1_mapped_address_ip);
     debug!("[{}] test1 is_same_ip: l:{} r:{}", stun_server, local_ip, test1_mapped_address_ip);
     if test1_is_same_ip {
         // no nat
-        debug!("[{}] test2 send: {}", stun_server, test2_message);
-        let result = single_send(stun_server, test2_message, mut_socket_ref).await;
+        debug!("[{}] test2 send: {:?}", stun_server, test2_message);
+        let result = single_send::<ChangeRequest, MappedAddress>(stun_server, test2_message, mut_socket_ref).await;
         debug!("[{}] test2: {}", stun_server, result.is_ok());
         if result.is_err() {
             return IoResult::Ok((stun_server_string, public_address,OpenInternet));
         } else {
-            debug!("[{}] test2 recv: {}", stun_server, result.unwrap());
+            debug!("[{}] test2 recv: {:?}", stun_server, result.unwrap());
             return IoResult::Ok((stun_server_string, public_address, SymmetricUdpFirewall));
         }
     } else {
         // nat
-        debug!("[{}] test2 send: {}", stun_server, test2_message);
-        let result = single_send(stun_server, test2_message, mut_socket_ref).await;
+        debug!("[{}] test2 send: {:?}", stun_server, test2_message);
+        let result = single_send::<ChangeRequest, MappedAddress>(stun_server, test2_message, mut_socket_ref).await;
         debug!("[{}] test2: {}", stun_server,result.is_ok());
         if result.is_ok() {
             let test2_message = result.unwrap();
-            debug!("[{}] test2 recv: {}", stun_server, test2_message);
+            debug!("[{}] test2 recv: {:?}", stun_server, test2_message);
             return IoResult::Ok((stun_server_string, public_address, FullCone));
         } else {
             // test1(2)
             let test1_address = test1_changed_address.to_string();
 
-            let test12_message = build_request_bind_message(transaction_id);
-            debug!("[{}] test12 send: {}", stun_server,test12_message);
+            let test12_message: Message<UnknownAttributes> = build_request_bind_message(transaction_id);
+            debug!("[{}] test12 send: {:?}", stun_server,test12_message);
             let result = single_send(
                 test1_address.as_str(),
                 test12_message,
@@ -367,28 +365,29 @@ pub async fn nat_detect(local_address: SocketAddr,stun_server: &str) -> IoResult
                 return IoResult::Ok((stun_server_string,public_address, Unknown));
             } else {
                 // Symmetric NAT
-                let test12_response: Message = result.unwrap();
-                debug!("[{}] test12 recv: {}", stun_server, test12_response);
-                let mut mapped_address = MappedAddress::default();
-                let _ = mapped_address.get_from_as(&test12_response, ATTR_MAPPED_ADDRESS);
-                let test12_mapped_address = mapped_address;
+                let test12_response: Message<stun_codec_blazh::rfc5389::Attribute> = result.unwrap();
+                debug!("[{}] test12 recv: {:?}", stun_server, test12_response);
+
+                let test12_mapped_address ={
+                    let opt: Option<&MappedAddress> = test12_response.get_attribute();
+                    match opt {
+                        None => return other_error(),
+                        Some(a) => a.address()
+                    }
+                };
                 debug!("[{}] test12 mapped_address: {}", stun_server,test12_mapped_address);
 
-                if !mut_socket_ref.local_addr()?.ip().eq(&test12_mapped_address.ip)
-                    && mut_socket_ref.local_addr()?.port() ==  test12_mapped_address.port
+                if !mut_socket_ref.local_addr()?.ip().eq(&test12_mapped_address.ip())
+                    && mut_socket_ref.local_addr()?.port() ==  test12_mapped_address.port()
                 {
                     return IoResult::Ok((stun_server_string, public_address,Symmetric));
                 } else {
                     // test 3
                     let test3_message = build_request_bind_message_with_attribute(
-                        transaction_id, RawAttribute{
-                            typ: ATTR_CHANGE_REQUEST,
-                            value: vec![0, 0, 0, 0b00000010],
-                            length: 4
-                        }
+                        transaction_id, ChangeRequest::new(false, true)
                     );
-                    debug!("[{}] test3 send: {}", stun_server, test3_message);
-                    let result = single_send(
+                    debug!("[{}] test3 send: {:?}", stun_server, test3_message);
+                    let result = single_send::<ChangeRequest, stun_codec_blazh::rfc5389::Attribute>(
                         test1_address.as_str(),
                         test3_message,
                         mut_socket_ref
@@ -397,7 +396,7 @@ pub async fn nat_detect(local_address: SocketAddr,stun_server: &str) -> IoResult
                     if result.is_err() {
                         return IoResult::Ok((stun_server_string,public_address, RestrictedCone));
                     } else {
-                        debug!("[{}] test3 recv: {}", stun_server, result.unwrap());
+                        debug!("[{}] test3 recv: {:?}", stun_server, result.unwrap());
                         return IoResult::Ok((stun_server_string, public_address,PortRestrictedCone));
                     }
                 }
@@ -412,29 +411,43 @@ fn other_error<A>() -> IoResult<A> {
     IoResult::Err(std::io::Error::from(ErrorKind::Other))
 }
 
-fn build_request_bind_message(transaction_id: TransactionId) -> Message {
-    let mut message = Message::new();
-    message.transaction_id = transaction_id;
-    message.typ = MessageType::new(
-        METHOD_BINDING,
-        MessageClass::default()
+fn build_request_bind_message<A: Attribute>(transaction_id: TransactionId) -> Message<A> {
+    return  Message::new(
+        MessageClass::Request,
+        methods::BINDING,
+        transaction_id
     );
-    message
 }
 
-fn build_request_bind_message_with_attribute(
-    transaction_id: TransactionId, a: RawAttribute
-) -> Message {
+fn build_request_bind_message_with_attribute<A: Attribute>(
+    transaction_id: TransactionId, a: A
+) -> Message<A> {
     let mut message = build_request_bind_message(transaction_id);
-    message.attributes.0.push(a);
+    message.add_attribute(a);
     message
 }
 
-async fn single_send(stun_server: &str, mut message: Message, socket: & mut UdpSocket)
-    -> IoResult<Message>
+// pub struct StunResult{
+//     request_id: TransactionId,
+//     bs: Vec<u8>
+// }
+//
+//
+// impl StunResult {
+//     fn decode<T: Attribute>(&self) -> IoResult<Message<T>> {
+//         todo!()
+//     }
+// }
+
+async fn single_send<A: Attribute, B: Attribute>(
+    stun_server: &str,
+    message: Message<A>,
+    socket: & mut UdpSocket
+)
+    -> IoResult<Message<B>>
 {
-    message.encode();
-    let bytes = message.raw;
+    let mut encoder = MessageEncoder::default();
+    let bytes: Vec<u8> = encoder.encode_into_bytes(message.clone()).map_err(|_e| std::io::Error::from(ErrorKind::Other))?;
     let mut buf = [0; 1 << 9];
     for _i in 0..STUN_RETRY_COUNT {
         match tokio::time::timeout(TIMEOUT, socket.send_to(bytes.as_slice(), stun_server)).await {
@@ -451,23 +464,21 @@ async fn single_send(stun_server: &str, mut message: Message, socket: & mut UdpS
                 }
             }
         };
-        let mut result_message = Message::new();
-        result_message.raw = (&buf[0..len]).to_vec();
-        if result_message.decode().is_err() {
-            // FIXME
-            break
-        }
-        if result_message.typ.class == CLASS_ERROR_RESPONSE {
-            // FIXME
-            break
-        }
-        if message.transaction_id.eq(&result_message.transaction_id) {
-            return IoResult::Ok(result_message);
-        } else {
-            // FIXME
-            break
-        }
 
+        let mut decoder = MessageDecoder::<B>::new();
+        match decoder.decode_from_bytes(&buf[0..len]) {
+            Ok(Ok(m)) => {
+                if m.class()== MessageClass::ErrorResponse {
+                    break
+                }
+                if message.transaction_id().eq(&m.transaction_id()) {
+                    return IoResult::Ok(m);
+                }
+            }
+            _ => {
+                continue
+            }
+        };
     }
     return IoResult::Err(std::io::Error::from(ErrorKind::Other));
 }
